#include <assert.h>
#include <fcntl.h>
#include <intrin.h>
#include <io.h>
#include <stdio.h>
#include "basetypes.h"
#include "Bios.h"
#include "byteswap.h"
#include "EmitMisc.h"
#include "EmitECU.h"
#include "EmitALU.h"
#include "EmitMEM.h"
#include "EmitMUL.h"
#include "EmitRCU.h"
#include "ExecuteECU.h"
#include "ExecuteRCU.h"
#include "ExecuteALU.h"
#include "ExecuteMUL.h"
#include "ExecuteMEM.h"
#include "ExecuteMisc.h"
#include "InstructionCache.h"
#include "InstructionDependencies.h"
#include "mpe.h"
#include "NativeCodeCache.h"
#include "NuonEnvironment.h"
#include "NuonMemoryMap.h"
#include "OverlayManager.h"
#include "PerformanceTimer.h"
#include "PresentationEngine.h"
#include "PrintECU.h"
#include "PrintRCU.h"
#include "PrintALU.h"
#include "PrintMUL.h"
#include "PrintMEM.h"
#include "PrintMisc.h"
#include "SuperBlock.h"
#include "X86EmitTypes.h"
#include "Utility.h"

//#define LOG_COMM
//#define LOG_PROGRAM_FLOW
//#define LOG_ADDRESS_ONLY
//#define LOG_BIOS_CALLS

#define LOG_MPE_INDEX (1)

#ifdef LOG_PROGRAM_FLOW
#define LOG_STUFF
#else 
#ifdef LOG_BIOS_CALLS
#define LOG_STUFF
#endif
#endif

extern NuonEnvironment *nuonEnv;
extern NuonBiosHandler BiosJumpTable[];
extern NuonBiosHandler DVDJumpTable[];
extern char *BiosRoutineNames[];
extern bool bCallingMediaCallback;
extern FILE *commLogFile;

uint8 MPE::mirrorLookup[256];

NuanceHandler nuanceHandlers[] =
{
  //ECU Executes
  Execute_ECU_NOP,
  Execute_Halt,
  Execute_BRAAlways,
  Execute_BRAAlways_NOP,
  Execute_BRAConditional,
  Execute_BRAConditional_NOP,
  Execute_JMPAlwaysIndirect,
  Execute_JMPAlwaysIndirect_NOP,
  Execute_JMPConditionalIndirect,
  Execute_JMPConditionalIndirect_NOP,
  Execute_JSRAlways,
  Execute_JSRAlways_NOP,
  Execute_JSRConditional,
  Execute_JSRConditional_NOP,
  Execute_JSRAlwaysIndirect,
  Execute_JSRAlwaysIndirect_NOP,
  Execute_JSRConditionalIndirect,
  Execute_JSRConditionalIndirect_NOP,
  Execute_RTI1Conditional,
  Execute_RTI1Conditional_NOP,
  Execute_RTI2Conditional,
  Execute_RTI2Conditional_NOP,
  Execute_RTSAlways,
  Execute_RTSAlways_NOP,
  Execute_RTSConditional,
  Execute_RTSConditional_NOP,
  //RCU Executes
  Execute_DECRc1,
  Execute_DECRc0,
  Execute_DECBoth,
  Execute_ADDRImmediateOnly,
  Execute_ADDRImmediate,
  Execute_ADDRScalarOnly,
  Execute_ADDRScalar,
  Execute_MVRImmediateOnly,
  Execute_MVRImmediate,
  Execute_MVRScalarOnly,
  Execute_MVRScalar,
  Execute_RangeOnly,
  Execute_Range,
  Execute_ModuloOnly,
  Execute_Modulo,
  //ALU Executes
  Execute_ABS,
  Execute_BITSScalar,
  Execute_BITSImmediate,
  Execute_BTST,
  Execute_BUTT,
  Execute_COPY,
  Execute_MSB,
  Execute_SAT,
  Execute_AS,
  Execute_ASL,
  Execute_ASR,
  Execute_LS,
  Execute_LSR,
  Execute_ROT,
  Execute_ROL,
  Execute_ROR,
  Execute_ADD_P,
  Execute_SUB_P,
  Execute_ADD_SV,
  Execute_SUB_SV,
  Execute_ADDImmediate,
  Execute_ADDScalar,
  Execute_ADDScalarShiftRightImmediate,
  Execute_ADDScalarShiftLeftImmediate,
  Execute_SUBImmediate,
  Execute_SUBImmediateReverse,
  Execute_SUBScalar,
  Execute_SUBScalarShiftRightImmediate,
  Execute_SUBScalarShiftLeftImmediate,
  Execute_CMPImmediate,
  Execute_CMPImmediateReverse,
  Execute_CMPScalar,
  Execute_CMPScalarShiftRightImmediate,
  Execute_CMPScalarShiftLeftImmediate,
  Execute_ANDImmediate,
  Execute_ANDScalar,
  Execute_ANDImmediateShiftScalar,
  Execute_ANDScalarShiftRightImmediate,
  Execute_ANDScalarShiftLeftImmediate,
  Execute_ANDScalarShiftScalar,
  Execute_ANDScalarRotateScalar,
  Execute_FTSTImmediate,
  Execute_FTSTScalar,
  Execute_FTSTImmediateShiftScalar,
  Execute_FTSTScalarShiftRightImmediate,
  Execute_FTSTScalarShiftLeftImmediate,
  Execute_FTSTScalarShiftScalar,
  Execute_FTSTScalarRotateScalar,
  Execute_ORImmediate,
  Execute_ORScalar,
  Execute_ORImmediateShiftScalar,
  Execute_ORScalarShiftRightImmediate,
  Execute_ORScalarShiftLeftImmediate,
  Execute_ORScalarShiftScalar,
  Execute_ORScalarRotateScalar,
  Execute_EORImmediate,
  Execute_EORScalar,
  Execute_EORImmediateShiftScalar,
  Execute_EORScalarShiftRightImmediate,
  Execute_EORScalarShiftLeftImmediate,
  Execute_EORScalarShiftScalar,
  Execute_EORScalarRotateScalar,
  Execute_ADDWCScalar,
  Execute_ADDWCImmediate,
  Execute_ADDWCScalarShiftRightImmediate,
  Execute_ADDWCScalarShiftLeftImmediate,
  Execute_SUBWCScalar,
  Execute_SUBWCImmediate,
  Execute_SUBWCImmediateReverse,
  Execute_SUBWCScalarShiftRightImmediate,
  Execute_SUBWCScalarShiftLeftImmediate,
  Execute_CMPWCScalar,
  Execute_CMPWCImmediate,
  Execute_CMPWCImmediateReverse,
  Execute_CMPWCScalarShiftRightImmediate,
  Execute_CMPWCScalarShiftLeftImmediate,
  //MUL Executes
  Execute_ADDM,
  Execute_ADDMImmediate,
  Execute_SUBM,
  Execute_SUBMImmediateReverse,
  Execute_MULScalarShiftAcshift,
  Execute_MULScalarShiftRightImmediate,
  Execute_MULScalarShiftLeftImmediate,
  Execute_MULImmediateShiftAcshift,
  Execute_MULScalarShiftScalar,
  Execute_MULImmediateShiftScalar,
  Execute_MULImmediateShiftRightImmediate,
  Execute_MULImmediateShiftLeftImmediate,
  Execute_MUL_SVImmediateShiftImmediate,
  Execute_MUL_SVScalarShiftImmediate,
  Execute_MUL_SVScalarShiftSvshift,
  Execute_MUL_SVRuShiftImmediate,
  Execute_MUL_SVRuShiftSvshift,
  Execute_MUL_SVRvShiftImmediate,
  Execute_MUL_SVRvShiftSvshift,
  Execute_MUL_SVVectorShiftImmediate,
  Execute_MUL_SVVectorShiftSvshift,
  Execute_MUL_PImmediateShiftImmediate,
  Execute_MUL_PScalarShiftImmediate,
  Execute_MUL_PScalarShiftSvshift,
  Execute_MUL_PRuShiftImmediate,
  Execute_MUL_PRuShiftSvshift,
  Execute_MUL_PRvShiftImmediate,
  Execute_MUL_PRvShiftSvshift,
  Execute_MUL_PVectorShiftImmediate,
  Execute_MUL_PVectorShiftSvshift,
  Execute_DOTPScalarShiftImmediate,
  Execute_DOTPScalarShiftSvshift,
  Execute_DOTPVectorShiftImmediate,
  Execute_DOTPVectorShiftSvshift,
  //MEM Executes
  Execute_Mirror,
  Execute_MV_SImmediate,
  Execute_MV_SScalar,
  Execute_MV_V,
  Execute_PopVector,
  Execute_PopVectorRz,
  Execute_PopScalarRzi1,
  Execute_PopScalarRzi2,
  Execute_PushVector,
  Execute_PushVectorRz,
  Execute_PushScalarRzi1,
  Execute_PushScalarRzi2,
  Execute_LoadScalarLinear,
  Execute_LoadScalarControlRegisterAbsolute,
  Execute_LoadByteAbsolute,
  Execute_LoadWordAbsolute,
  Execute_LoadScalarAbsolute,
  Execute_LoadShortVectorAbsolute,
  Execute_LoadVectorAbsolute,
  Execute_LoadVectorControlRegisterAbsolute,
  Execute_LoadPixelAbsolute,
  Execute_LoadPixelZAbsolute,
  Execute_LoadByteLinear,
  Execute_LoadByteBilinearUV,
  Execute_LoadByteBilinearXY,
  Execute_LoadWordLinear,
  Execute_LoadWordBilinearUV,
  Execute_LoadWordBilinearXY,
  Execute_LoadScalarBilinearUV,
  Execute_LoadScalarBilinearXY,
  Execute_LoadShortVectorLinear,
  Execute_LoadShortVectorBilinearUV,
  Execute_LoadShortVectorBilinearXY,
  Execute_LoadVectorLinear,
  Execute_LoadVectorBilinearUV,
  Execute_LoadVectorBilinearXY,
  Execute_LoadPixelLinear,
  Execute_LoadPixelBilinearUV,
  Execute_LoadPixelBilinearXY,
  Execute_LoadPixelZLinear,
  Execute_LoadPixelZBilinearUV,
  Execute_LoadPixelZBilinearXY,
  Execute_StoreScalarAbsolute,
  Execute_StoreScalarControlRegisterAbsolute,
  Execute_StoreShortVectorAbsolute,
  Execute_StoreVectorAbsolute,
  Execute_StoreVectorControlRegisterAbsolute,
  Execute_StorePixelAbsolute,
  Execute_StorePixelZAbsolute,
  Execute_StoreScalarLinear,
  Execute_StoreScalarBilinearUV,
  Execute_StoreScalarBilinearXY,
  Execute_StoreShortVectorLinear,
  Execute_StoreShortVectorBilinearUV,
  Execute_StoreShortVectorBilinearXY,
  Execute_StoreVectorLinear,
  Execute_StoreVectorBilinearUV,
  Execute_StoreVectorBilinearXY,
  Execute_StorePixelLinear,
  Execute_StorePixelBilinearUV,
  Execute_StorePixelBilinearXY,
  Execute_StorePixelZLinear,
  Execute_StorePixelZBilinearUV,
  Execute_StorePixelZBilinearXY,
  Execute_StoreScalarControlRegisterImmediate,
  Execute_StoreScalarImmediate,
  //MISC Executes
  Execute_StoreScalarRegisterConstant,
  Execute_StoreMiscRegisterConstant,
  Execute_ECU_NOP,//Packet_Start
  Execute_ECU_NOP,//Packet_End
  Execute_CheckECUSkipCounter, //Packet_End for packets with branches or packets in delay slots (IL block only)
  Execute_SaveFlags,
  Execute_SaveRegs,
};

NuancePrintHandler printHandlers[] =
{
  //ECU HANDLERS
  Print_ECU_NOP,
  Print_Halt,
  Print_BRAAlways,
  Print_BRAAlways_NOP,
  Print_BRAConditional,
  Print_BRAConditional_NOP,
  Print_JMPAlwaysIndirect,
  Print_JMPAlwaysIndirect_NOP,
  Print_JMPConditionalIndirect,
  Print_JMPConditionalIndirect_NOP,
  Print_JSRAlways,
  Print_JSRAlways_NOP,
  Print_JSRConditional,
  Print_JSRConditional_NOP,
  Print_JSRAlwaysIndirect,
  Print_JSRAlwaysIndirect_NOP,
  Print_JSRConditionalIndirect,
  Print_JSRConditionalIndirect_NOP,
  Print_RTI1Conditional,
  Print_RTI1Conditional_NOP,
  Print_RTI2Conditional,
  Print_RTI2Conditional_NOP,
  Print_RTSAlways,
  Print_RTSAlways_NOP,
  Print_RTSConditional,
  Print_RTSConditional_NOP,
  //RCU HANDLERS
  Print_DEC,
  Print_DEC,
  Print_DEC,
  Print_ADDRImmediate,
  Print_ADDRImmediate,
  Print_ADDRScalar,
  Print_ADDRScalar,
  Print_MVRImmediate,
  Print_MVRImmediate,
  Print_MVRScalar,
  Print_MVRScalar,
  Print_Range,
  Print_Range,
  Print_Modulo,
  Print_Modulo,
  //ALU HANDLERS
  Print_ABS,
  Print_BITSScalar,
  Print_BITSImmediate,
  Print_BTST,
  Print_BUTT,
  Print_COPY,
  Print_MSB,
  Print_SAT,
  Print_AS,
  Print_ASL,
  Print_ASR,
  Print_LS,
  Print_LSR,
  Print_ROT,
  Print_ROL,
  Print_ROR,
  Print_ADD_P,
  Print_SUB_P,
  Print_ADD_SV,
  Print_SUB_SV,
  Print_ADDImmediate,
  Print_ADDScalar,
  Print_ADDScalarShiftRightImmediate,
  Print_ADDScalarShiftLeftImmediate,
  Print_SUBImmediate,
  Print_SUBImmediateReverse,
  Print_SUBScalar,
  Print_SUBScalarShiftRightImmediate,
  Print_SUBScalarShiftLeftImmediate,
  Print_CMPImmediate,
  Print_CMPImmediateReverse,
  Print_CMPScalar,
  Print_CMPScalarShiftRightImmediate,
  Print_CMPScalarShiftLeftImmediate,
  Print_ANDImmediate,
  Print_ANDScalar,
  Print_ANDImmediateShiftScalar,
  Print_ANDScalarShiftRightImmediate,
  Print_ANDScalarShiftLeftImmediate,
  Print_ANDScalarShiftScalar,
  Print_ANDScalarRotateScalar,
  Print_FTSTImmediate,
  Print_FTSTScalar,
  Print_FTSTImmediateShiftScalar,
  Print_FTSTScalarShiftRightImmediate,
  Print_FTSTScalarShiftLeftImmediate,
  Print_FTSTScalarShiftScalar,
  Print_FTSTScalarRotateScalar,
  Print_ORImmediate,
  Print_ORScalar,
  Print_ORImmediateShiftScalar,
  Print_ORScalarShiftRightImmediate,
  Print_ORScalarShiftLeftImmediate,
  Print_ORScalarShiftScalar,
  Print_ORScalarRotateScalar,
  Print_EORImmediate,
  Print_EORScalar,
  Print_EORImmediateShiftScalar,
  Print_EORScalarShiftRightImmediate,
  Print_EORScalarShiftLeftImmediate,
  Print_EORScalarShiftScalar,
  Print_EORScalarRotateScalar,
  Print_ADDWCScalar,
  Print_ADDWCImmediate,
  Print_ADDWCScalarShiftRightImmediate,
  Print_ADDWCScalarShiftLeftImmediate,
  Print_SUBWCScalar,
  Print_SUBWCImmediate,
  Print_SUBWCImmediateReverse,
  Print_SUBWCScalarShiftRightImmediate,
  Print_SUBWCScalarShiftLeftImmediate,
  Print_CMPWCScalar,
  Print_CMPWCImmediate,
  Print_CMPWCImmediateReverse,
  Print_CMPWCScalarShiftRightImmediate,
  Print_CMPWCScalarShiftLeftImmediate,
  //MUL HANDLERS
  Print_ADDM,
  Print_ADDMImmediate,
  Print_SUBM,
  Print_SUBMImmediateReverse,
  Print_MULScalarShiftAcshift,
  Print_MULScalarShiftRightImmediate,
  Print_MULScalarShiftLeftImmediate,
  Print_MULImmediateShiftAcshift,
  Print_MULScalarShiftScalar,
  Print_MULImmediateShiftScalar,
  Print_MULImmediateShiftRightImmediate,
  Print_MULImmediateShiftLeftImmediate,
  Print_MUL_SVImmediateShiftImmediate,
  Print_MUL_SVScalarShiftImmediate,
  Print_MUL_SVScalarShiftSvshift,
  Print_MUL_SVRuShiftImmediate,
  Print_MUL_SVRuShiftSvshift,
  Print_MUL_SVRvShiftImmediate,
  Print_MUL_SVRvShiftSvshift,
  Print_MUL_SVVectorShiftImmediate,
  Print_MUL_SVVectorShiftSvshift,
  Print_MUL_PImmediateShiftImmediate,
  Print_MUL_PScalarShiftImmediate,
  Print_MUL_PScalarShiftSvshift,
  Print_MUL_PRuShiftImmediate,
  Print_MUL_PRuShiftSvshift,
  Print_MUL_PRvShiftImmediate,
  Print_MUL_PRvShiftSvshift,
  Print_MUL_PVectorShiftImmediate,
  Print_MUL_PVectorShiftSvshift,
  Print_DOTPScalarShiftImmediate,
  Print_DOTPScalarShiftSvshift,
  Print_DOTPVectorShiftImmediate,
  Print_DOTPVectorShiftSvshift,
  //MEM HANDLERS
  Print_Mirror,
  Print_MV_SImmediate,
  Print_MV_SScalar,
  Print_MV_V,
  Print_PopVector,
  Print_PopVectorRz,
  Print_PopScalarRzi1,
  Print_PopScalarRzi2,
  Print_PushVector,
  Print_PushVectorRz,
  Print_PushScalarRzi1,
  Print_PushScalarRzi2,
  Print_LoadScalarLinear,
  Print_LoadScalarControlRegisterAbsolute,
  Print_LoadByteAbsolute,
  Print_LoadWordAbsolute,
  Print_LoadScalarAbsolute,
  Print_LoadShortVectorAbsolute,
  Print_LoadVectorAbsolute,
  Print_LoadVectorControlRegisterAbsolute,
  Print_LoadPixelAbsolute,
  Print_LoadPixelZAbsolute,
  Print_LoadByteLinear,
  Print_LoadByteBilinearUV,
  Print_LoadByteBilinearXY,
  Print_LoadWordLinear,
  Print_LoadWordBilinearUV,
  Print_LoadWordBilinearXY,
  Print_LoadScalarBilinearUV,
  Print_LoadScalarBilinearXY,
  Print_LoadShortVectorLinear,
  Print_LoadShortVectorBilinearUV,
  Print_LoadShortVectorBilinearXY,
  Print_LoadVectorLinear,
  Print_LoadVectorBilinearUV,
  Print_LoadVectorBilinearXY,
  Print_LoadPixelLinear,
  Print_LoadPixelBilinearUV,
  Print_LoadPixelBilinearXY,
  Print_LoadPixelZLinear,
  Print_LoadPixelZBilinearUV,
  Print_LoadPixelZBilinearXY,
  Print_StoreScalarAbsolute,
  Print_StoreScalarControlRegisterAbsolute,
  Print_StoreShortVectorAbsolute,
  Print_StoreVectorAbsolute,
  Print_StoreVectorControlRegisterAbsolute,
  Print_StorePixelAbsolute,
  Print_StorePixelZAbsolute,
  Print_StoreScalarLinear,
  Print_StoreScalarBilinearUV,
  Print_StoreScalarBilinearXY,
  Print_StoreShortVectorLinear,
  Print_StoreShortVectorBilinearUV,
  Print_StoreShortVectorBilinearXY,
  Print_StoreVectorLinear,
  Print_StoreVectorBilinearUV,
  Print_StoreVectorBilinearXY,
  Print_StorePixelLinear,
  Print_StorePixelBilinearUV,
  Print_StorePixelBilinearXY,
  Print_StorePixelZLinear,
  Print_StorePixelZBilinearUV,
  Print_StorePixelZBilinearXY,
  Print_StoreScalarControlRegisterImmediate,
  Print_StoreScalarImmediate,
  //Misc
  Print_StoreScalarRegisterConstant,
  Print_StoreMiscRegisterConstant,
  Print_PacketStart,
  Print_PacketEnd,
  Print_CheckECUSkipCounter,
  Print_SaveFlags,
  Print_SaveRegs,
};

NativeEmitHandler emitHandlers[] =
{
  //ECU Executes
  Emit_NOP,           //Execute_ECU_NOP
  0,                  //Execute_Halt
  Emit_BRAAlways,
  Emit_BRAAlways_NOP,
  Emit_BRAConditional,
  Emit_BRAConditional_NOP,
  Emit_JMPAlwaysIndirect,
  Emit_JMPAlwaysIndirect_NOP,
  Emit_JMPConditionalIndirect,
  Emit_JMPConditionalIndirect_NOP,
  Emit_JSRAlways,
  Emit_JSRAlways_NOP,
  Emit_JSRConditional,
  Emit_JSRConditional_NOP,
  Emit_JSRAlwaysIndirect,
  Emit_JSRAlwaysIndirect_NOP,
  Emit_JSRConditionalIndirect,
  Emit_JSRConditionalIndirect_NOP,
  Emit_RTI1Conditional,
  Emit_RTI1Conditional_NOP,
  Emit_RTI2Conditional,
  Emit_RTI2Conditional_NOP,
  Emit_RTSAlways,
  Emit_RTSAlways_NOP,
  Emit_RTSConditional,
  Emit_RTSConditional_NOP,
  //RCU Executes
  Emit_DECRc1,
  Emit_DECRc0,
  Emit_DECBoth,
  Emit_ADDRImmediateOnly,
  Emit_ADDRImmediate,    
  Emit_ADDRScalarOnly,   
  Emit_ADDRScalar,       
  Emit_MVRImmediateOnly, 
  Emit_MVRImmediate,     
  Emit_MVRScalarOnly,    
  Emit_MVRScalar,        
  Emit_RangeOnly,        
  Emit_Range,            
  Emit_ModuloOnly,       
  Emit_Modulo,           
  //ALU Executes
  Emit_ABS,
  Emit_BITSScalar,
  Emit_BITSImmediate,
  Emit_BTST,
  Emit_BUTT,
  Emit_COPY,
  Emit_MSB,
  Emit_SAT,
  Emit_AS,
  Emit_ASL,
  Emit_ASR,
  Emit_LS,
  Emit_LSR,
  Emit_ROT,
  Emit_ROL,
  Emit_ROR,
  Emit_ADD_P,
  Emit_SUB_P,
  Emit_ADD_SV,
  Emit_SUB_SV,
  Emit_ADDImmediate,
  Emit_ADDScalar,
  Emit_ADDScalarShiftRightImmediate,
  Emit_ADDScalarShiftLeftImmediate,
  Emit_SUBImmediate,
  Emit_SUBImmediateReverse,
  Emit_SUBScalar,
  Emit_SUBScalarShiftRightImmediate,
  Emit_SUBScalarShiftLeftImmediate,
  Emit_CMPImmediate,
  Emit_CMPImmediateReverse,
  Emit_CMPScalar,
  Emit_CMPScalarShiftRightImmediate,
  Emit_CMPScalarShiftLeftImmediate,
  Emit_ANDImmediate,
  Emit_ANDScalar,
  Emit_ANDImmediateShiftScalar,
  Emit_ANDScalarShiftRightImmediate,
  Emit_ANDScalarShiftLeftImmediate,
  Emit_ANDScalarShiftScalar,
  Emit_ANDScalarRotateScalar,
  Emit_FTSTImmediate,
  Emit_FTSTScalar,
  Emit_FTSTImmediateShiftScalar,
  Emit_FTSTScalarShiftRightImmediate,
  Emit_FTSTScalarShiftLeftImmediate,
  Emit_FTSTScalarShiftScalar,
  Emit_FTSTScalarRotateScalar,
  Emit_ORImmediate,
  Emit_ORScalar,
  Emit_ORImmediateShiftScalar,
  Emit_ORScalarShiftRightImmediate,
  Emit_ORScalarShiftLeftImmediate,
  Emit_ORScalarShiftScalar,
  Emit_ORScalarRotateScalar,
  Emit_EORImmediate,
  Emit_EORScalar,
  Emit_EORImmediateShiftScalar,
  Emit_EORScalarShiftRightImmediate,
  Emit_EORScalarShiftLeftImmediate,
  Emit_EORScalarShiftScalar,
  Emit_EORScalarRotateScalar,
  Emit_ADDWCScalar,
  Emit_ADDWCImmediate,
  Emit_ADDWCScalarShiftRightImmediate,
  Emit_ADDWCScalarShiftLeftImmediate,
  Emit_SUBWCScalar,
  Emit_SUBWCImmediate,
  Emit_SUBWCImmediateReverse,
  Emit_SUBWCScalarShiftRightImmediate,
  Emit_SUBWCScalarShiftLeftImmediate,
  Emit_CMPWCScalar,
  Emit_CMPWCImmediate,
  Emit_CMPWCImmediateReverse,
  Emit_CMPWCScalarShiftRightImmediate,
  Emit_CMPWCScalarShiftLeftImmediate,
  //MUL Executes
  Emit_ADDM,                    
  Emit_ADDMImmediate,                    
  Emit_SUBM,                    
  Emit_SUBMImmediateReverse,                    
  
  Emit_MULScalarShiftAcshift,                   
  Emit_MULScalarShiftRightImmediate,
  Emit_MULScalarShiftLeftImmediate,
  Emit_MULImmediateShiftAcshift,
  Emit_MULScalarShiftScalar,
  Emit_MULImmediateShiftScalar,
  Emit_MULImmediateShiftRightImmediate,
  Emit_MULImmediateShiftLeftImmediate,

  Emit_MUL_SVImmediateShiftImmediate,
  Emit_MUL_SVScalarShiftImmediate,
  Emit_MUL_SVScalarShiftSvshift,
  Emit_MUL_SVRuShiftImmediate,
  Emit_MUL_SVRuShiftSvshift,
  Emit_MUL_SVRvShiftImmediate,
  Emit_MUL_SVRvShiftSvshift,
  Emit_MUL_SVVectorShiftImmediate,
  Emit_MUL_SVVectorShiftSvshift,
  Emit_MUL_PImmediateShiftImmediate,
  Emit_MUL_PScalarShiftImmediate,
  Emit_MUL_PScalarShiftSvshift,
  Emit_MUL_PRuShiftImmediate,
  Emit_MUL_PRuShiftSvshift,
  Emit_MUL_PRvShiftImmediate,
  Emit_MUL_PRvShiftSvshift,
  Emit_MUL_PVectorShiftImmediate,
  Emit_MUL_PVectorShiftSvshift,
  Emit_DOTPScalarShiftImmediate,
  Emit_DOTPScalarShiftSvshift,
  Emit_DOTPVectorShiftImmediate,
  Emit_DOTPVectorShiftSvshift,
  //MEM Executes
  Emit_Mirror,
  Emit_MV_SImmediate,
  Emit_MV_SScalar,
  Emit_MV_V,
  Emit_PopVector,
  Emit_PopVectorRz,
  Emit_PopScalarRzi1,
  Emit_PopScalarRzi2,
  Emit_PushVector,
  Emit_PushVectorRz,
  Emit_PushScalarRzi1,
  Emit_PushScalarRzi2,
  Emit_LoadScalarLinear,
  Emit_LoadScalarControlRegisterAbsolute,
  Emit_LoadByteAbsolute,
  Emit_LoadWordAbsolute,
  Emit_LoadScalarAbsolute,
  Emit_LoadShortVectorAbsolute,
  Emit_LoadVectorAbsolute,
  0,                    //Execute_LoadVectorControlRegisterAbsolute,
  Emit_LoadPixelAbsolute,
  Emit_LoadPixelZAbsolute,
  Emit_LoadByteLinear,
  Emit_LoadByteBilinearUV,
  Emit_LoadByteBilinearXY,
  Emit_LoadWordLinear,
  Emit_LoadWordBilinearUV,
  Emit_LoadWordBilinearXY,
  Emit_LoadScalarBilinearUV,
  Emit_LoadScalarBilinearXY,
  Emit_LoadShortVectorLinear,
  Emit_LoadShortVectorBilinearUV,
  Emit_LoadShortVectorBilinearXY,
  Emit_LoadVectorLinear,
  Emit_LoadVectorBilinearUV,
  Emit_LoadVectorBilinearXY,
  Emit_LoadPixelLinear,
  Emit_LoadPixelBilinearUV,
  Emit_LoadPixelBilinearXY,
  Emit_LoadPixelZLinear,
  Emit_LoadPixelZBilinearUV,
  Emit_LoadPixelZBilinearXY,
  Emit_StoreScalarAbsolute,
  Emit_StoreScalarControlRegisterAbsolute,
  Emit_StoreShortVectorAbsolute,
  Emit_StoreVectorAbsolute,
  0,                    //Execute_StoreVectorControlRegisterAbsolute,
  Emit_StorePixelAbsolute,
  Emit_StorePixelZAbsolute,
  Emit_StoreScalarLinear,
  Emit_StoreScalarBilinearUV,
  Emit_StoreScalarBilinearXY,
  Emit_StoreShortVectorLinear,
  Emit_StoreShortVectorBilinearUV,
  Emit_StoreShortVectorBilinearXY,
  Emit_StoreVectorLinear,
  Emit_StoreVectorBilinearUV,
  Emit_StoreVectorBilinearXY,
  Emit_StorePixelLinear,
  Emit_StorePixelBilinearUV,
  Emit_StorePixelBilinearXY,
  Emit_StorePixelZLinear,
  Emit_StorePixelZBilinearUV,
  Emit_StorePixelZBilinearXY,
  Emit_StoreScalarControlRegisterImmediate,
  Emit_StoreScalarImmediate,
  //MISC Executes
  Emit_StoreScalarRegisterConstant,
  Emit_StoreMiscRegisterConstant,
  Emit_NOP,            //Packet_Start
  Emit_NOP,            //Packet_End
  Emit_NOP,            //Packet_End (CheckECUSkipCounter)
  Emit_SaveRegs,
  Emit_SaveRegs,
};

FILE *logfile = NULL;
double timeDelta;
uint64 timer_start, timer_end;

MPE::MPE(uint32 index)
{
  EmitterVariables emitvars;
  uint32 numCacheEntries[] = {4096,2048,2048,262144};
  uint32 numTLBEntries[] = {4096,2048,2048,98304};
  uint32 overlayLengths[] = {8192,4096,4096,4096};

  numInterpreterCacheFlushes = 0;
  numNativeCodeCacheFlushes = 0;
  numNonCompilablePackets = 0;
  mpeIndex = index;
  bStrictMemoryDependencyPolicy = true;
  strictMemoryMiscInputDependencies = 0xFFFFFFFF;
  strictMemoryMiscOutputDependencies = 0xFFFFFFFF;
  AllocateMPELocalMemory();
  GenerateMirrorLookupTable();
  nativeCodeCache = new NativeCodeCache(5UL*1024UL*1024UL, 0, numTLBEntries[mpeIndex & 0x03]);
  instructionCache = new InstructionCache(numCacheEntries[mpeIndex & 0x03]);
  overlayManager = new OverlayManager();
  overlayManager->SetOverlayLength(overlayLengths[mpeIndex]);
  bInvalidateInstructionCaches = false;
  bInvalidateInterpreterCache = false;
  
  interpretNextPacket = 0;

  emitvars.mpe = this;
  emitvars.codeCache = nativeCodeCache;
  emitvars.regBase = (uint32)&cc;
  emitvars.tempRegBase = (uint32)&tempCC;
  emitvars.ppEmitLoc = nativeCodeCache->GetEmitPointerAddress();
  emitvars.patchMgr = nativeCodeCache->patchMgr;
  nativeCodeCache->SetEmitVars(emitvars);

  //SuperBlock(mpe,maxPackets, maxInstructionsPerPacket)
  superBlock = new SuperBlock(this,120,5);
  InitStaticICacheEntries();
  if(mpeIndex == LOG_MPE_INDEX)
  {      
    if(!logfile)
    {
      logfile = fopen("logfile","w");
    }

    if(!commLogFile)
    {
      commLogFile = fopen("commlog","w");
    }
  }
  Reset();
}

MPE::~MPE()
{
  delete nativeCodeCache;
  delete instructionCache;
  delete overlayManager;
  delete superBlock;
  if(mpeIndex == LOG_MPE_INDEX)
  {
    if(logfile)
    {
      fclose(logfile);
    }

    if(commLogFile)
    {
      fclose(commLogFile);
    }
  }
  FreeMPELocalMemory();
}

void MPE::AllocateMPELocalMemory()
{
  dtrom = new uint8[MPE_LOCAL_MEMORY_SIZE];
}

void MPE::FreeMPELocalMemory()
{
  delete [] dtrom;
}

void MPE::GenerateMirrorLookupTable()
{
  uint8 mirror;

  for(uint32 i = 0; i <= 0xFF; i++)
  {
    mirror = 0;
    mirror |= ((i & 0x01) << 7) & 0x80;
    mirror |= ((i & 0x02) << 5) & 0x40;
    mirror |= ((i & 0x04) << 3) & 0x20;
    mirror |= ((i & 0x08) << 1) & 0x10;
    mirror |= ((i & 0x10) >> 1) & 0x08;
    mirror |= ((i & 0x20) >> 3) & 0x04;
    mirror |= ((i & 0x40) >> 5) & 0x02;
    mirror |= ((i & 0x80) >> 7) & 0x01;

    mirrorLookup[i] = mirror;
  }
}

void MPE::InitializeBankTable(uint8 *mainBusPtr, uint8 *systemBusPtr, uint8 *flashEEPROMPtr)
{
  //Initialize the bank pointer lookup table for use with indirect
  //loads and stores.  This table assumes that the only banks present
  //are local MPE memory, system bus memory and main bus memory.  All
  //other banks map to main bus memory in the hope that a bad memory address
  //write will simply cause main bus graphics corruption

  //This table cannot distinguish between the ROM, reserved and other bus IO
  //banks in the $F0000000/$F1000000/$FE000000 range but programs should
  //never read or write in this range anyways.  If these banks are emulated
  //at some future point, this table will need to expand to 256 entries in
  //order to factor in the entire upper byte of the 32 bit address range

  bankPtrTable[0x0] = dtrom;
  bankPtrTable[0x1] = dtrom;
  bankPtrTable[0x2] = dtrom;
  bankPtrTable[0x3] = mainBusPtr;
  bankPtrTable[0x4] = mainBusPtr;
  bankPtrTable[0x5] = mainBusPtr;
  bankPtrTable[0x6] = mainBusPtr;
  bankPtrTable[0x7] = mainBusPtr;
  bankPtrTable[0x8] = systemBusPtr;
  bankPtrTable[0x9] = systemBusPtr;
  bankPtrTable[0xA] = systemBusPtr;
  bankPtrTable[0xB] = systemBusPtr;
  bankPtrTable[0xC] = systemBusPtr;
  bankPtrTable[0xD] = systemBusPtr;
  bankPtrTable[0xE] = systemBusPtr;
  bankPtrTable[0xF] = flashEEPROMPtr;
}

const uint32 COMPILE_THRESHOLD = 50UL;

void MPE::Reset()
{
  //Explicitly clear CSTATE bits to zero so that code works in debug mode
  dcachectl &= ~0xF0000000;
  icachectl &= ~0xF0000000;

  instructionCache->Invalidate();
  nativeCodeCache->Flush();

  invalidateRegionStart = MPE_IRAM_BASE;
  invalidateRegionEnd = (MPE_IRAM_BASE + OVERLAY_SIZE - 1);
  interpreterInvalidateRegionStart = 0;
  interpreterInvalidateRegionEnd = 0;

  overlayIndex = 0;

  //Interpretation of Nuances require the use of the cc composite flags register 
  bUsingCompositeFlags = true;
  ecuSkipCounter = 0;
  pcfetchnext = 0x20300000;
  pcfetch = 0x20300000;
  pcroute = 0x20300000;
  pcexec = 0x20300000;
  sp = 0x20001000;
  //mpectl is halted on reset and has the was-reset bit set
  mpectl = MPECTRL_MPEWASRESET;
  //all exceptions halt the processor
  excephalten = 0x1FFF;
  //reset all exceptions
  excepsrc = 0;
  //intctl is set to 0 on reset according to SDK docs
  intctl = 0;
  //reset all interrupt sources
  intsrc = 0;
  //reset interupt enable registers
  inten1 = 0;
  //set level 2 selector to software
  inten2sel = 1;
  //enable other bus dma, no commands pending, no other bus activity
  odmactl = 1UL << 5;
  //set bus priority for MDMA transfers to 3 and clear all other bits
  mdmactl = 3UL << 5;
  commxmit0 = 0;
  commxmit1 = 0;
  commxmit2 = 0;
  commxmit3 = 0;
  commrecv0 = 0;
  commrecv1 = 0;
  commrecv2 = 0;
  commrecv3 = 0;
  acshift = 0;
  svshift = 0;
  commctl = 0;
  rc0 = 0;
  rc1 = 0;
  cc = (CC_COUNTER0_ZERO | CC_COUNTER1_ZERO);
  cycleCounter = 0;
  //Nuon = aries 2, MPE release = aries 2, mpe identifier, halted on reset
  configa = (3UL << 24) | (3UL << 16) | (mpeIndex << 8) | 0;
  configb = 0;
}

bool MPE::LoadBinaryFile(uchar *filename, bool bIRAM)
{
  int handle;
  int byteLength;
  uint8 *buffer = &dtrom[MPE_IROM_OFFSET];

  handle = open((char *)filename,_O_RDONLY|_O_BINARY,0);
  if(handle >= 0)
  {
    byteLength = filelength(handle);

    if(bIRAM)
    {
      buffer = &dtrom[MPE_IRAM_OFFSET];
    }

    read(handle, buffer, byteLength);
    close(handle);
    return true;
  }
  else
  {
    return false;
  }
}

inline bool MPE::ExecuteUntilAddress(uint32 address)
{
  bool status;
  
  breakpointAddress = address;
  status = FetchDecodeExecute();
  breakpointAddress = 0;

  return status;
}

inline uint32 MPE::GetPacketDelta(uint8 *iPtr, uint32 numLevels)
{
  uint8 opcode;
  uint8 deltaBytes, packetBytes;
  bool bTerminating;

  packetBytes = 0;
  deltaBytes = 0;

  while(numLevels != 0)
  {
    opcode = *iPtr;
    if(opcode <= 0x3F)
    {
      bTerminating = true;
      deltaBytes += 2;
      packetBytes += 2;
      iPtr += 2;
    }
    else
    {
      //Instruction is not a 16 bit ALU instruction

      //If 16 bit non-terminating non-ALU instruction, control instruction
      //or 32 bit immediate extension
      if(opcode < 0x88 || opcode > 0x9F)
      {
        //If not a control instruction (PAD, NOP or BREAKPOINT)
        if((opcode & 0xFC) != 0x80)
        {
          //if bit 15 is set, the instruction is a packet terminator
          bTerminating = (opcode & 0x80);
        }
        else
        {
          //if bit 8 is set, the instruction is a terminating breakpoint or NOP
          bTerminating = (opcode & 0x01);
        }

        deltaBytes += 2;
        packetBytes += 2;
        iPtr += 2;
      }
      else if(opcode >= 0x90)
      {
        //32 bit instruction: if bit 12 is set, instruction is a packet terminator
        bTerminating = *(iPtr + 2) & 0x10;
        deltaBytes += 4;
        packetBytes += 4;
        iPtr += 4;
      }
      else
      {
        //32 bit extension (preceeding 48/64 bit instruction)
        bTerminating = false;
        deltaBytes += 4;
        packetBytes += 4;
        iPtr += 4;
      }
    }

    if(bTerminating || (packetBytes >= 128))
    {
      numLevels--;
      packetBytes = 0;
    }
  }

  return deltaBytes;
}

void MPE::DecompressPacket(uint8 *iBuffer, InstructionCacheEntry *pICacheEntry, uint32 options)
{
  InstructionCacheEntry pStruct;
  uint32 executionUnits, i;
  uint32 ecuIndex, rcuIndex, aluIndex, mulIndex, memIndex, baseIndex;
  uint32 immExt = 0;
  uint32 packetByteCount = 0;
  uint32 comboScalarInDep;
  uint32 comboMiscInDep;
  uint32 comboScalarOutDep;
  uint32 comboMiscOutDep;
  bool bTerminating = false;
  bool bCanEmitNativeCode = true;
  uint8 iLength;

  pStruct.pcexec = pICacheEntry->pcexec;
  pICacheEntry->pcroute = pICacheEntry->pcexec;
  pICacheEntry->nuanceCount = 0;
  pStruct.packetInfo = 0;

  for(i = 0; i < 5; i++)
  {
    pStruct.scalarInputDependencies[i] = 0;
    pStruct.miscInputDependencies[i] = 0;
    pStruct.scalarOutputDependencies[i] = 0;
    pStruct.miscOutputDependencies[i] = 0;
    pICacheEntry->scalarInputDependencies[i] = 0;
    pICacheEntry->miscInputDependencies[i] = 0;
    pICacheEntry->scalarOutputDependencies[i] = 0;
    pICacheEntry->miscOutputDependencies[i] = 0;
  }

  do
  {
    iLength = DecodeSingleInstruction(iBuffer,&pStruct,&immExt,bTerminating);

    pICacheEntry->pcroute += iLength;

    if(packetByteCount >= 128)
    {
      pICacheEntry->pcroute = pICacheEntry->pcexec + 128;
      break;
    }

    iBuffer += iLength;
  }
  while(!bTerminating);

  pStruct.nuances[FIXED_FIELD(SLOT_ECU,FIELD_ECU_PCROUTE)] = pICacheEntry->pcroute;

  if(pStruct.packetInfo & PACKETINFO_NEEDS_PCFETCHNEXT)
  {
    pICacheEntry->pcfetchnext = pICacheEntry->pcroute + GetPacketDelta(iBuffer, 2);
    pStruct.nuances[FIXED_FIELD(SLOT_ECU,FIELD_ECU_PCFETCHNEXT)] = pICacheEntry->pcfetchnext;
  }

  if(options & DECOMPRESS_OPTIONS_INHIBIT_ECU)
  {
    pStruct.packetInfo &= ~(PACKETINFO_ECU | PACKETINFO_BRANCH_CONDITIONAL | PACKETINFO_BRANCH_ALWAYS | PACKETINFO_BRANCH_NOP);
  }

  executionUnits = GETPACKETEXECUTIONUNITS(pStruct.packetInfo);
 
  switch(executionUnits)
  {
    case 0:
      pICacheEntry->nuanceCount = 0;
      break;
    case (PACKETINFO_ECU >> 2):
      pICacheEntry->nuanceCount = 1;
      pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
      break;
    case (PACKETINFO_RCU >> 2):
      pICacheEntry->nuanceCount = 1;
      pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_RCU);
      break;
    case ((PACKETINFO_RCU | PACKETINFO_ECU) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_RCU);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
      }
      else
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
      }    
      break;
    case (PACKETINFO_MEM >> 2):
      pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
      pICacheEntry->nuanceCount = 1;
      break;
    case ((PACKETINFO_MEM | PACKETINFO_ECU) >> 2): 
      pICacheEntry->nuanceCount = 2;
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_ECU,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
      {
        memIndex = 0;
        ecuIndex = 1;
      }
      else
      {
        memIndex = 1;
        ecuIndex = 0;
      }       
      pICacheEntry->CopyInstructionData(ecuIndex,&pStruct,SLOT_ECU);
      pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
      break;
    case ((PACKETINFO_MEM | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 2;
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_RCU,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
      {
        memIndex = 0;
        rcuIndex = 1;
      }
      else
      {
        memIndex = 1;
        rcuIndex = 0;
      }       
      pICacheEntry->CopyInstructionData(rcuIndex,&pStruct,SLOT_RCU);
      pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
      break;
    case ((PACKETINFO_MEM | PACKETINFO_ECU | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(ChooseInstructionPairOrdering(&pStruct,SLOT_RCU,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
        {
          memIndex = 0;
          rcuIndex = 1;
        }
        else
        {
          memIndex = 1;
          rcuIndex = 0;
        }        
        pICacheEntry->CopyInstructionData(rcuIndex,&pStruct,SLOT_RCU);
        pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
        }
        else
        {
          ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_MEM,SLOT_ECU,SLOT_RCU);
        }
      }
      break;
    case (PACKETINFO_MUL >> 2):
      pICacheEntry->nuanceCount = 1;
      pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MUL);
      break;
    case ((PACKETINFO_MUL | PACKETINFO_ECU) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MUL);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
      }
      else
      {
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MUL);
      }
      break;
    case ((PACKETINFO_MUL | PACKETINFO_RCU) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_RCU,SLOT_MUL))
      {
        mulIndex = 0;
        rcuIndex = 1;
      }
      else
      {
        mulIndex = 1;
        rcuIndex = 0;
      }
      pICacheEntry->CopyInstructionData(rcuIndex,&pStruct,SLOT_RCU);
      pICacheEntry->CopyInstructionData(mulIndex,&pStruct,SLOT_MUL);
      break;
    case ((PACKETINFO_MUL | PACKETINFO_RCU | PACKETINFO_ECU) >> 2): 
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        baseIndex = 0;
        ecuIndex = 2;
      }
      else
      {
        baseIndex = 1;
        ecuIndex = 0;
      }

      if(ChooseInstructionPairOrdering(&pStruct,SLOT_RCU,SLOT_MUL))
      {
        mulIndex = baseIndex + 0;
        rcuIndex = baseIndex + 1;
      }
      else
      {
        mulIndex = baseIndex + 1;
        rcuIndex = baseIndex + 0;
      }
      
      pICacheEntry->CopyInstructionData(ecuIndex,&pStruct,SLOT_ECU);
      pICacheEntry->CopyInstructionData(rcuIndex,&pStruct,SLOT_RCU);
      pICacheEntry->CopyInstructionData(mulIndex,&pStruct,SLOT_MUL);
      break;
    case ((PACKETINFO_MUL | PACKETINFO_MEM) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_MUL,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
      {
        memIndex = 0;
        mulIndex = 1;
      }
      else
      {
        memIndex = 1;
        mulIndex = 0;
      }       
      pICacheEntry->CopyInstructionData(mulIndex,&pStruct,SLOT_MUL);
      pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
      break;
    case ((PACKETINFO_MUL | PACKETINFO_MEM | PACKETINFO_ECU) >> 2): 
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(ChooseInstructionPairOrdering(&pStruct,SLOT_MUL,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
        {
          memIndex = 0;
          mulIndex = 1;
        }
        else
        {
          memIndex = 1;
          mulIndex = 0;
        }       
        
        pICacheEntry->CopyInstructionData(mulIndex,&pStruct,SLOT_MUL);
        pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_MUL);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
        }
        else
        {
          ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_MUL,SLOT_MEM,SLOT_ECU);
        }
      }
      break;
    case ((PACKETINFO_MUL | PACKETINFO_MEM | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_MUL);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_RCU);
      }
      else
      {
        ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_MUL,SLOT_MEM,SLOT_RCU);
      }
      break;
    case ((PACKETINFO_MUL | PACKETINFO_MEM | PACKETINFO_ECU | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 4;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_MUL);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        }
        else
        {
          ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_MUL,SLOT_MEM,SLOT_RCU);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        }
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_MUL);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        }
        else
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
          ScheduleInstructionTriplet(pICacheEntry,1,&pStruct,SLOT_MUL,SLOT_MEM,SLOT_RCU);
        }
      }
      break;
    case (PACKETINFO_ALU >> 2):
      pICacheEntry->nuanceCount = 1;
      pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ALU);
      break;
    case ((PACKETINFO_ALU | PACKETINFO_ECU) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ALU);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
      }
      else
      {
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ALU);
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_RCU) >> 2):
      pICacheEntry->nuanceCount = 2;
      pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_RCU);
      pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ALU);
      break;
    case ((PACKETINFO_ALU | PACKETINFO_RCU | PACKETINFO_ECU) >> 2):
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_RCU);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ALU);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
      }
      else
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ALU);
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MEM) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_ALU,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
      {
        memIndex = 0;
        aluIndex = 1;
      }
      else
      {
        memIndex = 1;
        aluIndex = 0;
      }       
      pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
      pICacheEntry->CopyInstructionData(aluIndex,&pStruct,SLOT_ALU);
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MEM | PACKETINFO_ECU) >> 2): 
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(ChooseInstructionPairOrdering(&pStruct,SLOT_ALU,SLOT_MEM) || (options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST))
        {
          memIndex = 0;
          aluIndex = 1;
        }
        else
        {
          memIndex = 1;
          aluIndex = 0;
        }       
        pICacheEntry->CopyInstructionData(memIndex,&pStruct,SLOT_MEM);
        pICacheEntry->CopyInstructionData(aluIndex,&pStruct,SLOT_ALU);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ALU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ECU);
        }
        else
        {
          ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MEM,SLOT_ECU);
        }
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MEM | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ALU);
      }
      else
      {
        ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MEM,SLOT_RCU);
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MEM | PACKETINFO_ECU | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 4;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ALU);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        }
        else
        {
          ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MEM,SLOT_RCU);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        }
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {    
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_RCU);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ALU);
        }
        else
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
          ScheduleInstructionTriplet(pICacheEntry,1,&pStruct,SLOT_ALU,SLOT_MEM,SLOT_RCU);
        }
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL) >> 2):
      pICacheEntry->nuanceCount = 2;
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_ALU,SLOT_MUL))
      {
        mulIndex = 0;
        aluIndex = 1;
      }
      else
      {
        mulIndex = 1;
        aluIndex = 0;
      }
      pICacheEntry->CopyInstructionData(mulIndex,&pStruct,SLOT_MUL);
      pICacheEntry->CopyInstructionData(aluIndex,&pStruct,SLOT_ALU);
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_ECU) >> 2):
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        ecuIndex = 2;
        baseIndex = 0;
      }
      else
      {
        ecuIndex = 0;
        baseIndex = 1;
      }
      
      if(ChooseInstructionPairOrdering(&pStruct,SLOT_ALU,SLOT_MUL))
      {
        mulIndex = baseIndex + 0;
        aluIndex = baseIndex + 1;
      }
      else
      {
        mulIndex = baseIndex + 1;
        aluIndex = baseIndex + 0;
      }
      pICacheEntry->CopyInstructionData(ecuIndex,&pStruct,SLOT_ECU);
      pICacheEntry->CopyInstructionData(mulIndex,&pStruct,SLOT_MUL);
      pICacheEntry->CopyInstructionData(aluIndex,&pStruct,SLOT_ALU);
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_RCU) >> 2):
      pICacheEntry->nuanceCount = 3;
      ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_RCU);
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_RCU | PACKETINFO_ECU) >> 2):
      pICacheEntry->nuanceCount = 4;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_RCU);
      }
      else
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
        ScheduleInstructionTriplet(pICacheEntry,1,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_RCU);
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_MEM) >> 2):
      pICacheEntry->nuanceCount = 3;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
        pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ALU);
        pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_MUL);
      }
      else
      {
        ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_MEM);
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_MEM | PACKETINFO_ECU) >> 2): 
      pICacheEntry->nuanceCount = 4;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ALU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_MUL);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
        }
        else
        {
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_ECU);
          ScheduleInstructionTriplet(pICacheEntry,0,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_MEM);
        }
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ALU);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_MUL);
        }
        else
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
          ScheduleInstructionTriplet(pICacheEntry,1,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_MEM);
        }
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_MEM | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 4;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
      {
        pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
        ScheduleInstructionTriplet(pICacheEntry,1,&pStruct,SLOT_ALU,SLOT_MUL,SLOT_RCU);
      }
      else
      {
        ScheduleInstructionQuartet(pICacheEntry,0,&pStruct);
      }
      break;
    case ((PACKETINFO_ALU | PACKETINFO_MUL | PACKETINFO_MEM | PACKETINFO_ECU | PACKETINFO_RCU) >> 2): 
      pICacheEntry->nuanceCount = 5;
      if(options & DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST)
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_RCU);
          pICacheEntry->CopyInstructionData(2,&pStruct,SLOT_ALU);
          pICacheEntry->CopyInstructionData(3,&pStruct,SLOT_MUL);
          pICacheEntry->CopyInstructionData(4,&pStruct,SLOT_ECU);
        }
        else
        {
          pICacheEntry->CopyInstructionData(4,&pStruct,SLOT_ECU);
          ScheduleInstructionQuartet(pICacheEntry,0,&pStruct);
        }
      }
      else
      {
        if(options & DECOMPRESS_OPTIONS_SCHEDULE_MEM_FIRST)
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_MEM);
          pICacheEntry->CopyInstructionData(1,&pStruct,SLOT_ECU);
          ScheduleInstructionTriplet(pICacheEntry,2,&pStruct,SLOT_RCU,SLOT_ALU,SLOT_MUL);
        }
        else
        {
          pICacheEntry->CopyInstructionData(0,&pStruct,SLOT_ECU);
          ScheduleInstructionQuartet(pICacheEntry,1,&pStruct);
        }
      }
      break;
  }

  comboScalarInDep = 0;
  comboMiscInDep = 0;
  comboScalarOutDep = pICacheEntry->scalarOutputDependencies[0];
  comboMiscOutDep = pICacheEntry->miscOutputDependencies[0];

  for(i = 1; i < 5; i++)
  {
    comboScalarInDep |= (pICacheEntry->scalarInputDependencies[i] & comboScalarOutDep);
    comboMiscInDep |= (pICacheEntry->miscInputDependencies[i] & comboMiscOutDep);
    comboScalarOutDep |= pICacheEntry->scalarOutputDependencies[i];
    comboMiscOutDep |= pICacheEntry->miscOutputDependencies[i];
  }

  pICacheEntry->packetInfo = pStruct.packetInfo;
  pICacheEntry->ecuConditionCode = pStruct.ecuConditionCode;

  if(!nuonEnv->compilerOptions.bAllowCompile)
  {
    pICacheEntry->packetInfo |= PACKETINFO_NEVERCOMPILE;
  }

  if(comboScalarInDep | comboMiscInDep)
  {
    pICacheEntry->packetInfo |= PACKETINFO_DEPENDENCY_PRESENT;

    pICacheEntry->pScalarRegs = tempScalarRegs;
    pICacheEntry->pIndexRegs = &tempRx;
    pICacheEntry->pRzRegs = &tempRz;
    pICacheEntry->pCounterRegs = &tempRc0;
    pICacheEntry->pAcshift = &tempAcshift;
    pICacheEntry->pSvshift = &tempSvshift;
    pICacheEntry->pXyctl = &tempXyctl;
    pICacheEntry->pUvctl = &tempUvctl;
    pICacheEntry->pXyrange = &tempXyrange;
    pICacheEntry->pUvrange = &tempUvrange;
  }
  else
  {
    pICacheEntry->pScalarRegs = regs;
    pICacheEntry->pIndexRegs = &rx;
    pICacheEntry->pRzRegs = &rz;
    pICacheEntry->pCounterRegs = &rc0;
    pICacheEntry->pAcshift = &acshift;
    pICacheEntry->pSvshift = &svshift;
    pICacheEntry->pXyctl = &xyctl;
    pICacheEntry->pUvctl = &uvctl;
    pICacheEntry->pXyrange = &xyrange;
    pICacheEntry->pUvrange = &uvrange;
  }

  pICacheEntry->handlers[0] = pICacheEntry->nuances[0];
  pICacheEntry->handlers[1] = pICacheEntry->nuances[5];
  pICacheEntry->handlers[2] = pICacheEntry->nuances[10];
  pICacheEntry->handlers[3] = pICacheEntry->nuances[15];
  pICacheEntry->handlers[4] = pICacheEntry->nuances[20];
}

inline bool MPE::ChooseInstructionPairOrdering(InstructionCacheEntry *entry, uint32 slot1, uint32 slot2)
{
  uint32 xScalarInDep, yScalarInDep, xMiscInDep, yMiscInDep;
  uint32 xScalarOutDep, yScalarOutDep, xMiscOutDep, yMiscOutDep;

  xScalarInDep = entry->scalarInputDependencies[slot1];
  yScalarInDep = entry->scalarInputDependencies[slot2];
  xMiscInDep = entry->miscInputDependencies[slot1];
  yMiscInDep = entry->miscInputDependencies[slot2];
  xScalarOutDep = entry->scalarOutputDependencies[slot1];
  yScalarOutDep = entry->scalarOutputDependencies[slot2];
  xMiscOutDep = entry->miscOutputDependencies[slot1];
  yMiscOutDep = entry->miscOutputDependencies[slot2];
  return (OnesCount(yScalarInDep & xScalarOutDep) + OnesCount(yMiscInDep & xMiscOutDep)) 
    > (OnesCount(xScalarInDep & yScalarOutDep) + OnesCount(xMiscInDep & yMiscOutDep));
}

void MPE::ScheduleInstructionTriplet(InstructionCacheEntry *destEntry, uint32 baseSlot, InstructionCacheEntry *srcEntry, uint32 slot1, uint32 slot2, uint32 slot3)
{
  static uint32 destSlot1[6] = {0,0,1,1,2,2};
  static uint32 destSlot2[6] = {1,2,2,0,0,1};
  static uint32 destSlot3[6] = {2,1,0,2,1,0};

  uint32 comboScalarOutDep12, comboScalarOutDep23, comboScalarOutDep13;
  uint32 comboMiscOutDep12, comboMiscOutDep23, comboMiscOutDep13;
  uint32 tempScalarInDep, tempMiscInDep, i, minVal, minIndex;
  uint32 scores[6];
  
/*
  After completion, scores will contain the dependency count for the following permuations:
  {[1,2,3],[1,3,2],[2,3,1],[2,1,3],[3,1,2],[3,2,1]}.  Choose the permutation with the lowest
  score.
*/

  comboScalarOutDep12 = srcEntry->scalarOutputDependencies[slot1] | srcEntry->scalarOutputDependencies[slot2];
  comboMiscOutDep12 = srcEntry->miscOutputDependencies[slot1] | srcEntry->miscOutputDependencies[slot2];

  tempScalarInDep = srcEntry->scalarInputDependencies[slot3] & comboScalarOutDep12;
  tempMiscInDep = srcEntry->miscInputDependencies[slot3] & comboMiscOutDep12;

  scores[0] = 
    OnesCount((srcEntry->scalarInputDependencies[slot2] & srcEntry->scalarOutputDependencies[slot1]) |
              (tempScalarInDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot2] & srcEntry->miscOutputDependencies[slot1]) |
              (tempMiscInDep));

  scores[3] = 
    OnesCount((srcEntry->scalarInputDependencies[slot1] & srcEntry->scalarOutputDependencies[slot2]) |
              (tempScalarInDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot1] & srcEntry->miscOutputDependencies[slot2]) |
              (tempMiscInDep));

  comboScalarOutDep13 = srcEntry->scalarOutputDependencies[slot1] | srcEntry->scalarOutputDependencies[slot3];
  comboMiscOutDep13 = srcEntry->miscOutputDependencies[slot1] | srcEntry->miscOutputDependencies[slot3];

  tempScalarInDep = srcEntry->scalarInputDependencies[slot2] & comboScalarOutDep13;
  tempMiscInDep = srcEntry->miscInputDependencies[slot2] & comboMiscOutDep13;

  scores[1] = 
    OnesCount((srcEntry->scalarInputDependencies[slot3] & srcEntry->scalarOutputDependencies[slot1]) |
              (tempScalarInDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot3] & srcEntry->miscOutputDependencies[slot1]) |
              (tempMiscInDep));

  scores[4] = 
    OnesCount((srcEntry->scalarInputDependencies[slot1] & srcEntry->scalarOutputDependencies[slot3]) |
              (tempScalarInDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot1] & srcEntry->miscOutputDependencies[slot3]) |
              (tempMiscInDep));

  comboScalarOutDep23 = srcEntry->scalarOutputDependencies[slot2] | srcEntry->scalarOutputDependencies[slot3];
  comboMiscOutDep23 = srcEntry->miscOutputDependencies[slot2] | srcEntry->miscOutputDependencies[slot3];

  tempScalarInDep = srcEntry->scalarInputDependencies[slot1] & comboScalarOutDep23;
  tempMiscInDep = srcEntry->miscInputDependencies[slot1] & comboMiscOutDep23;

  scores[2] = 
    OnesCount((srcEntry->scalarInputDependencies[slot3] & srcEntry->scalarOutputDependencies[slot2]) |
              (tempScalarInDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot3] & srcEntry->miscOutputDependencies[slot2]) |
              (tempMiscInDep));

  scores[5] = 
    OnesCount((srcEntry->scalarInputDependencies[slot2] & srcEntry->scalarOutputDependencies[slot3]) |
              (tempScalarInDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot2] & srcEntry->miscOutputDependencies[slot3]) |
              (tempMiscInDep));

  minVal = scores[0];
  minIndex = 0;

  for(i = 1; i < 6; i++)
  {
    if(scores[i] <= minVal)
    {
      minIndex = i;
      minVal = scores[i];
    }
  }

  destEntry->CopyInstructionData(baseSlot + destSlot1[minIndex], srcEntry, slot1);
  destEntry->CopyInstructionData(baseSlot + destSlot2[minIndex], srcEntry, slot2);
  destEntry->CopyInstructionData(baseSlot + destSlot3[minIndex], srcEntry, slot3);
}

uint32 MPE::ScoreInstructionTriplet(InstructionCacheEntry *srcEntry, uint32 slot1, uint32 slot2, uint32 slot3)
{
  uint32 comboScalarOutDep, comboMiscOutDep;

  comboMiscOutDep = srcEntry->miscOutputDependencies[slot1] | srcEntry->miscOutputDependencies[slot2];
  comboScalarOutDep = srcEntry->scalarOutputDependencies[slot1] | srcEntry->scalarOutputDependencies[slot2];

  return OnesCount((srcEntry->scalarInputDependencies[slot2] & srcEntry->scalarOutputDependencies[slot1]) |
    (srcEntry->scalarInputDependencies[slot3] & comboScalarOutDep)) +
    OnesCount((srcEntry->miscInputDependencies[slot2] & srcEntry->miscOutputDependencies[slot1]) |
    (srcEntry->miscInputDependencies[slot3] & comboMiscOutDep));
}

void MPE::GetInstructionTripletDependencies(uint32 *comboScalarDep, uint32 *comboMiscDep, InstructionCacheEntry *srcEntry, uint32 slot1, uint32 slot2, uint32 slot3)
{
  uint32 comboScalarOutDep, comboMiscOutDep;

  comboMiscOutDep = srcEntry->miscOutputDependencies[slot1] | srcEntry->miscOutputDependencies[slot2];
  comboScalarOutDep = srcEntry->scalarOutputDependencies[slot1] | srcEntry->scalarOutputDependencies[slot2];

  *comboMiscDep = (srcEntry->miscInputDependencies[slot2] & srcEntry->miscOutputDependencies[slot1]) |
    (srcEntry->miscInputDependencies[slot3] & comboMiscOutDep);
  *comboScalarDep = (srcEntry->scalarInputDependencies[slot2] & srcEntry->scalarOutputDependencies[slot1]) |
    (srcEntry->scalarInputDependencies[slot3] & comboScalarOutDep);
}

void MPE::ScheduleInstructionQuartet(InstructionCacheEntry *destEntry, uint32 baseSlot, InstructionCacheEntry *srcEntry)
{
  static uint32 destSlotRCU[6] = {0,0,1,1,2,0};
  static uint32 destSlotALU[6] = {1,2,2,3,3,3};
  static uint32 destSlotMUL[6] = {2,1,0,2,1,2};
  static uint32 destSlotMEM[6] = {3,3,3,0,0,1};

  uint32 comboScalarOutDep1, comboScalarOutDep2;
  uint32 comboMiscOutDep1, comboMiscOutDep2;
  uint32 tempScalarDep1, tempMiscDep1, tempScalarDep2, tempMiscDep2;
  uint32 minVal, minIndex, i;
  uint32 scores[6];

  tempScalarDep1 = srcEntry->scalarOutputDependencies[SLOT_RCU] | srcEntry->scalarOutputDependencies[SLOT_MUL];
  tempMiscDep1 = srcEntry->miscOutputDependencies[SLOT_RCU] | srcEntry->miscOutputDependencies[SLOT_MUL];
  
  comboScalarOutDep1 = srcEntry->scalarOutputDependencies[SLOT_ALU] | tempScalarDep1;
  comboMiscOutDep1 = srcEntry->miscOutputDependencies[SLOT_ALU] | tempMiscDep1;

  comboScalarOutDep2 = srcEntry->scalarOutputDependencies[SLOT_MEM] | tempScalarDep1;
  comboMiscOutDep2 = srcEntry->miscOutputDependencies[SLOT_MEM] | tempMiscDep1;

  tempScalarDep2 = srcEntry->scalarInputDependencies[SLOT_MEM] & comboScalarOutDep1;
  tempMiscDep2 = srcEntry->miscInputDependencies[SLOT_MEM] & comboMiscOutDep1;

  GetInstructionTripletDependencies(&tempScalarDep1,&tempMiscDep1,srcEntry,SLOT_RCU,SLOT_ALU,SLOT_MUL);
  scores[0] = OnesCount(tempScalarDep1 | tempScalarDep2) + OnesCount(tempMiscDep1 | tempMiscDep2);

  GetInstructionTripletDependencies(&tempScalarDep1,&tempMiscDep1,srcEntry,SLOT_RCU,SLOT_MUL,SLOT_ALU);
  scores[1] = OnesCount(tempScalarDep1 | tempScalarDep2) + OnesCount(tempMiscDep1 | tempMiscDep2);

  GetInstructionTripletDependencies(&tempScalarDep1,&tempMiscDep1,srcEntry,SLOT_MUL,SLOT_RCU,SLOT_ALU);
  scores[2] = OnesCount(tempScalarDep1 | tempScalarDep2) + OnesCount(tempMiscDep1 | tempMiscDep2);

  tempScalarDep2 = srcEntry->scalarInputDependencies[SLOT_ALU] & comboScalarOutDep2;
  tempMiscDep2 = srcEntry->miscInputDependencies[SLOT_ALU] & comboMiscOutDep2;

  GetInstructionTripletDependencies(&tempScalarDep1,&tempMiscDep1,srcEntry,SLOT_MEM,SLOT_RCU,SLOT_MUL);
  scores[3] = OnesCount(tempScalarDep1 | tempScalarDep2) + OnesCount(tempMiscDep1 | tempMiscDep2);

  GetInstructionTripletDependencies(&tempScalarDep1,&tempMiscDep1,srcEntry,SLOT_MEM,SLOT_MUL,SLOT_RCU);
  scores[4] = OnesCount(tempScalarDep1 | tempScalarDep2) + OnesCount(tempMiscDep1 | tempMiscDep2);
 
  GetInstructionTripletDependencies(&tempScalarDep1,&tempMiscDep1,srcEntry,SLOT_RCU,SLOT_MEM,SLOT_MUL);
  scores[5] = OnesCount(tempScalarDep1 | tempScalarDep2) + OnesCount(tempMiscDep1 | tempMiscDep2);

  minVal = scores[0];
  minIndex = 0;

  for(i = 1; i < 6; i++)
  {
    if(scores[i] <= minVal)
    {
      minIndex = i;
      minVal = scores[i];
    }
  }

  destEntry->CopyInstructionData(baseSlot + destSlotRCU[minIndex], srcEntry, SLOT_RCU);
  destEntry->CopyInstructionData(baseSlot + destSlotALU[minIndex], srcEntry, SLOT_ALU);
  destEntry->CopyInstructionData(baseSlot + destSlotMUL[minIndex], srcEntry, SLOT_MUL);
  destEntry->CopyInstructionData(baseSlot + destSlotMEM[minIndex], srcEntry, SLOT_MEM);
}

void LogMemoryLocation(FILE *outFile, char *varname, uint32 address, MPE *mpe)
{
  uint32 value = *((uint32 *)nuonEnv->GetPointerToMemory(mpe,address));
  SwapScalarBytes(&value);
  fprintf(outFile,"%s = $%8.8lx\n",varname,value);
}

void MPE::UpdateInvalidateRegion(uint32 start, uint32 length)
{
  uint32 end = (start + length - 1);

  invalidateRegionStart = (((start < invalidateRegionStart) || !invalidateRegionStart) ? start : invalidateRegionStart);
  invalidateRegionEnd = ((end > invalidateRegionEnd) ? end : invalidateRegionEnd);

  interpreterInvalidateRegionStart = (((start < interpreterInvalidateRegionStart) || !interpreterInvalidateRegionStart) ? start : interpreterInvalidateRegionStart);
  interpreterInvalidateRegionEnd = ((end > interpreterInvalidateRegionEnd) ? end : interpreterInvalidateRegionEnd);
}

bool MPE::FetchDecodeExecute()
{
  NativeCodeCacheEntry *pNativeCodeCacheEntry;
  InstructionCacheEntry *pInstructionCacheEntry;
  Nuance *pNuance;
  uint32 nInstructions, i, oldOverlayIndex, newOverlayIndex, pcexecLookupValue;
  bool bInvalidateOverlayRegion;
  bool bError;
  bool memory_valid;
  uint32 blockExecuteCount = 100;
  uint32 bCacheEntryValid;

  NativeCodeCacheEntryPoint nativeCodeCacheEntryPoint;

  cycleCounter = 0;

  if(mpectl & MPECTRL_MPEGO)
  {
    //StartPerformanceTimer();

    bInvalidateOverlayRegion = false;
    nativeCodeCacheEntryPoint = 0;
    pNativeCodeCacheEntry = 0;
    pInstructionCacheEntry = 0;

    /* Force 16 bit alignment of pcexec */
    pcexec &= ~0x01UL;

    /* Check for interrupts and update pcexec, rzi1 and rzi2 if an interrupt is to be serviced */

    if(intsrc && (ecuSkipCounter == 0))
    {
      //Test imaskHw2 bit
      if((intctl & (1UL << 5)) == 0)
      {
        //imaskHw2 not set
        //Test to see if the level 2 interrupt has occurred
        if(intsrc & (1UL << inten2sel))
        {
          //Test imaskSw2 mask
          if((intctl & (1UL << 7)) == 0)
          {
            //imaskSw2 not set so jump to the level 2 interrupt vector
            rzi2 = pcexec;
            pcexec = intvec2;
            //set imaskHw2 flag
            intctl |= (1UL << 5);
          }
        }
        else if((intctl & ((1UL << 3) | (1UL << 1))) == 0)
        {
          //imaskHw2 not set
          //neither imaskSw1 nor imaskHw1 set
          //Test to see if an enabled level 1 interrupt has occurred
          if(intsrc & inten1)
          {
            //Jump to the level 1 interrupt vector
            rzi1 = pcexec;
            pcexec = intvec1;
            //set imaskHw1 flag
            intctl |= (1UL << 1);
          }
        }
      }
    }

    pcexecLookupValue = pcexec;

    /* Now check to see if the MPE is executing out of local MPE memory.  If it is, then check to see if the MPE is executing */
    /* code within a region that has been marked for invalidation.  If executing with an invalidated region, the overlay manager */
    /* needs to hash the local MPE memory and see if the hash matches that of any existing compiled overlays in its list.  If it */
    /* can't find a match, a new overlay index is allocated or an existing set of compiled overlay code is replaced.  It a match is */
    /* made then that index is used.  Once an index is assigned, the index is combined with the current pcexec value to map the overlay */
    /* to an unused region of the Nuon address space to allow multiple sets of compiled overlays to co-exist even when the overlays */
    /* correspond to the same physical memory addresses.  As an example, the first set of code to be executed from MPE memory will be */
    /* assigned overlay index 0 and the compiled code will be assigned to the code cache range $20300000-$20307FFF.  If new overlay code
    /* is loaded into the MPE local memory and executed, the hash will not match that of overlay index 0 and so overlay index 1 will be */
    /* assigned with a code cache range of $20308000-$2030FFFF.  This is an important optimization as games like Tempest 3000 load */
    /* multiple overlays into the MPEs several times per frame.  If compiled overlay code was not allowed to exist, the overlay code would */
    /* require compiliation every time new overlay code was loaded into an MPE even when ping-ponging between two sets of overlay code. */
    /* The code cache entries would also require invalidation each time this happened. */

    if((pcexec < (MPE_IRAM_BASE + OVERLAY_SIZE)) && (pcexec >= MPE_IRAM_BASE))
    {
      //pcexec is within local MPE IRAM address space
      if((pcexec <= invalidateRegionEnd) && (pcexec >= invalidateRegionStart))
      {
        //pcexec is within MPE IRAM region that has been modified since the last time it was hashed

        overlayIndex = overlayManager->FindOverlay((uint32 *)&dtrom[MPE_IRAM_OFFSET], bInvalidateOverlayRegion);

        //Get the new overlay mask
        overlayMask = overlayManager->GetOverlayMask();

        //Invalidate the interpreter cache because these entries are not mapped to unique address ranges
        
        //numInterpreterCacheFlushes++;
        //instructionCache->InvalidateRegion(invalidateRegionStart, invalidateRegionEnd);
        bInvalidateInterpreterCache = true;

        if(bInvalidateOverlayRegion)
        {
          //The overlay manager assigned a previously used overlay ID so invalidate the code cache entries
          //associated with the overlay address range
          numNativeCodeCacheFlushes++;
          nativeCodeCache->FlushRegion(invalidateRegionStart | overlayMask, invalidateRegionEnd | overlayMask);
        }
ResetInvalidateRegion:
        //Reset the IRAM invalidation indicators
        invalidateRegionStart = 0xFFFFFFFFUL;
        invalidateRegionEnd = 0x00000000UL;
      }

      //Modify the pcexec lookup value.  If pcexec is not within MPE IRAM, the lookup value will remain equal to
      //pcexec otherwise it will be equal to pcexec ORed with the overlay mask

      pcexecLookupValue |= overlayMask;
    }
    else
    {
      if(bInvalidateInstructionCaches)
      {
        bInvalidateInstructionCaches = false;
        instructionCache->Invalidate();
        numInterpreterCacheFlushes++;
        numNativeCodeCacheFlushes++;
        if((mpeIndex == 0) || (mpeIndex == 3))
        {
          nativeCodeCache->FlushRegion(MAIN_BUS_BASE, MAIN_BUS_BASE + MAIN_BUS_SIZE - 1);
          nativeCodeCache->FlushRegion(SYSTEM_BUS_BASE, SYSTEM_BUS_BASE + SYSTEM_BUS_SIZE - 1);
        }
      }
    }

    if(!(ecuSkipCounter | interpretNextPacket))
    {
find_code_cache_entry:
      if(pNativeCodeCacheEntry = nativeCodeCache->FindNativeCodeCacheEntry(pcexecLookupValue))
      {
        nativeCodeCacheEntryPoint = pNativeCodeCacheEntry->entryPoint;
        goto execute_block;
      }
    
      if(bInvalidateInterpreterCache)
      {
        numInterpreterCacheFlushes++;
        instructionCache->InvalidateRegion(interpreterInvalidateRegionStart, interpreterInvalidateRegionEnd);
        bInvalidateInterpreterCache = false;
        interpreterInvalidateRegionStart = 0xFFFFFFFFUL;
        interpreterInvalidateRegionEnd = 0x00000000;
      }

      bCacheEntryValid = 1;
      pInstructionCacheEntry = instructionCache->FindInstructionCacheEntry(pcexec,bCacheEntryValid);
      if(bCacheEntryValid && (pcexec == pInstructionCacheEntry->pcexec))
      {
check_compile_threshhold:
        if(!(pInstructionCacheEntry->packetInfo & (PACKETINFO_COMPILED | PACKETINFO_NEVERCOMPILE)) && (pInstructionCacheEntry->frequencyCount >= COMPILE_THRESHOLD))
        {
          if(nativeCodeCache->IsBeyondThreshold())
          {
            numNativeCodeCacheFlushes++;
            nativeCodeCache->Flush();
            instructionCache->ClearCompiledStates();
          }

//#define COMPILE_TYPE SUPERBLOCKCOMPILETYPE_IL_BLOCK
#define COMPILE_TYPE SUPERBLOCKCOMPILETYPE_NATIVE_CODE_BLOCK

          nativeCodeCacheEntryPoint = CompileNativeCodeBlock(pcexecLookupValue, COMPILE_TYPE, bError);
          if(!bError)
          {
            pNativeCodeCacheEntry = nativeCodeCache->FindNativeCodeCacheEntry(pcexecLookupValue);
            if(nuonEnv->compilerOptions.bDumpBlocks)
            {
              //if(pNativeCodeCacheEntry->compileType == SUPERBLOCKCOMPILETYPE_NATIVE_CODE_BLOCK)
              //if(pNativeCodeCacheEntry->compileType == SUPERBLOCKCOMPILETYPE_IL_BLOCK)
              {
                superBlock->PrintBlockToFile(pNativeCodeCacheEntry->compileType, pNativeCodeCacheEntry->codeSize);
              }
            }

            pInstructionCacheEntry->packetInfo |= PACKETINFO_COMPILED;
            goto execute_block;
          }
          else
          {
            if(((int32)nativeCodeCacheEntryPoint) == -1)
            {
              pInstructionCacheEntry->packetInfo |= PACKETINFO_NEVERCOMPILE;
              numNonCompilablePackets++;
            }
            else if(nativeCodeCacheEntryPoint == 0)
            {
              numNativeCodeCacheFlushes++;
              nativeCodeCache->Flush();
              instructionCache->ClearCompiledStates();
            }

            nativeCodeCacheEntryPoint = 0;
            goto execute_block;
          }
        }
        else
        {
          pInstructionCacheEntry->frequencyCount++;
          if(!pInstructionCacheEntry->frequencyCount)
          {
            pInstructionCacheEntry->frequencyCount--;
          }
          goto execute_block;
        }
      }

      goto create_icache_entry;
    }
    else
    {
find_icache_entry:
      bCacheEntryValid = 1;
      pInstructionCacheEntry = instructionCache->FindInstructionCacheEntry(pcexec,bCacheEntryValid);
      if(bCacheEntryValid && (pcexec == pInstructionCacheEntry->pcexec))
      {
        goto execute_block;
      }

create_icache_entry:
      pInstructionCacheEntry->pcexec = pcexec;
      pInstructionCacheEntry->frequencyCount = 1;
      if(pcexec < ROM_BIOS_BASE)
      {
        DecompressPacket(GetPointerToMemoryBank(pcexec),pInstructionCacheEntry,DECOMPRESS_OPTIONS_SCHEDULE_ECU_LAST);
        if(((pcexec >= BIOS_JUMPTABLE_START) && (pcexec <= BIOS_JUMPTABLE_END))  || (pInstructionCacheEntry->packetInfo & PACKETINFO_BREAKPOINT))
        {
          pInstructionCacheEntry->packetInfo |= PACKETINFO_NEVERCOMPILE;            
        }
      }
      else
      {
        pInstructionCacheEntry->nuanceCount = 0;
        pInstructionCacheEntry->pcroute = 0;
        pInstructionCacheEntry->packetInfo |= PACKETINFO_NEVERCOMPILE;
      }
      instructionCache->SetEntryValid(pcexec);
execute_block:
      interpretNextPacket = 0;

    //StopPerformanceTimer();
    //timeDelta = GetTimeDeltaMs();

#ifdef LOG_STUFF
      if(LOG_MPE_INDEX == mpeIndex)
      {
        fprintf(logfile,"pcexec: %8.8x\n",pcexec);
      }
#endif

      if(nativeCodeCacheEntryPoint)
      {
        pNativeCodeCacheEntry->accessCount += 1;
        cycleCounter += pNativeCodeCacheEntry->numPackets;

        if(!pNativeCodeCacheEntry->accessCount)
        {
          pNativeCodeCacheEntry->accessCount -= 1;          
        }

        if((pNativeCodeCacheEntry->compileType == SUPERBLOCKCOMPILETYPE_IL_BLOCK) || (pNativeCodeCacheEntry->compileType == SUPERBLOCKCOMPILETYPE_IL_SINGLE))
        {
          nInstructions = pNativeCodeCacheEntry->numInstructions;
          pNuance = (Nuance *)nativeCodeCacheEntryPoint;
          bInterpretedBranchTaken = false;
          prevPcexec = pcexec;
          for(i = 0; i < nInstructions; i++)
          {
            (nuanceHandlers[pNuance->fields[0]])(*this,*pICacheEntry,*pNuance);
            pNuance++;

            if(bInterpretedBranchTaken)
            {
              pcexec = pcfetchnext;
              goto check_for_halt;
            }
          }
          pcexec = pNativeCodeCacheEntry->nextVirtualAddress;
          goto check_for_halt;
        }
        else
        {
          prevPcexec = pcexec;
          //do
          {
            nativeCodeCacheEntryPoint();
            //blockExecuteCount--;
          } 
          //while((pcexec == prevPcexec) && (mpectl & MPECTRL_MPEGO) && !ecuSkipCounter && blockExecuteCount);       
        }
      }
      else
      {
        if(pcexec < ROM_BIOS_BASE)
        {
          cycleCounter++;

          prevPcexec = pcexec;
          pcroute = pInstructionCacheEntry->pcroute;
          pcexec = pcroute;
          ExecuteNuances(*pInstructionCacheEntry);
        }
        else if(pcexec < 0xF0008000)
        {
          cycleCounter++;
          //Execute BIOS function: force to one of 256 entries
          BiosJumpTable[(pcexec >> 1) & 0xFF](this);

          if(!bCallingMediaCallback)
          {
            //Perform an implicit RTS, nop
            pcexec = rz;
          }
        }
        else
        {
          //Execute PE function
          cycleCounter++;
          CallPEHandler(this, pcexec);
          if(!bCallingMediaCallback)
          {
            //Perform an implicit RTS, nop
            pcexec = rz;
          }
        }
        
        bCallingMediaCallback = false;
      }

      if(ecuSkipCounter)
      {
        ecuSkipCounter--;
        if(!ecuSkipCounter)
        {
#ifdef LOG_BIOS_CALLS
          if(((pcfetchnext >= 0x80000000) && (pcfetchnext <= 0x8000FFFF)) && (mpeIndex == LOG_MPE_INDEX))
          {
            if(logfile)
            {
              if(pcfetchnext >= 0x80008000)
              {
#ifdef LOG_ADDRESS_ONLY
                fprintf(logfile,"PE CALL: $%8.8lX\n",pcfetchnext);
                fflush(logfile);
#else
                //fprintf(logfile,"PE CALL: %s ($%8.8lX): V0 = [$%lX,$%lX,$%lX,%lX], R4 = %lX, R5 = %lX\n",BiosRoutineNames[(pcfetchnext >> 3) & 0x3FFF],pcfetchnext,regs[0],regs[1],regs[2],regs[3],regs[4],regs[5]);
                fprintf(logfile,"PE CALL: %s ($%8.8lX)n",BiosRoutineNames[(pcfetchnext >> 3) & 0x3FFF],pcfetchnext,regs[0],regs[1],regs[2],regs[3],regs[4],regs[5]);
                fflush(logfile);
#endif
              }
              else
              {
#ifdef LOG_ADDRESS_ONLY
                fprintf(logfile,"BIOS CALL: %s ($%8.8lX)\n",BiosRoutineNames[(pcfetchnext >> 3) & 0x3FFF],pcfetchnext);
                fflush(logfile);
#else
                fprintf(logfile,"BIOS CALL: %s ($%8.8lX): V0 = [$%lX,$%lX,$%lX,%lX], R4 = %lX, R5 = %lX\n",BiosRoutineNames[(pcfetchnext >> 3) & 0x3FFF],pcfetchnext,regs[0],regs[1],regs[2],regs[3],regs[4],regs[5]);
                fflush(logfile);
#endif
              }
            }
          }
#endif
          pcexec = pcfetchnext;
        }
      }
      
check_for_halt:

      if(excephalten & excepsrc)
      {
        Halt();
      }
      else if(pcexec == breakpointAddress)
      {
        mpectl &= ~MPECTRL_MPEGO;
      }

    }
    
    //StopPerformanceTimer();
    //timeDelta = GetTimeDeltaMs();
    return true;
  }

  //StopPerformanceTimer();
  //timeDelta = GetTimeDeltaMs();
  return false;
}

void MPE::InitStaticICacheEntries()
{
  ICacheEntry_SaveRegs.pScalarRegs = tempScalarRegs;
  ICacheEntry_SaveRegs.pIndexRegs = &tempRx;
  ICacheEntry_SaveRegs.pRzRegs = &tempRz;
  ICacheEntry_SaveRegs.pCounterRegs = &tempRc0;
  ICacheEntry_SaveRegs.pAcshift = &tempAcshift;
  ICacheEntry_SaveRegs.pSvshift = &tempSvshift;
  ICacheEntry_SaveRegs.pXyctl = &tempXyctl;
  ICacheEntry_SaveRegs.pUvctl = &tempUvctl;
  ICacheEntry_SaveRegs.pXyrange = &tempXyrange;
  ICacheEntry_SaveRegs.pUvrange = &tempUvrange;
  
  ICacheEntry_SaveFlags.pScalarRegs = regs;
  ICacheEntry_SaveFlags.pIndexRegs = &rx;
  ICacheEntry_SaveFlags.pRzRegs = &rz;
  ICacheEntry_SaveFlags.pCounterRegs = &rc0;
  ICacheEntry_SaveFlags.pAcshift = &acshift;
  ICacheEntry_SaveFlags.pSvshift = &svshift;
  ICacheEntry_SaveFlags.pXyctl = &xyctl;
  ICacheEntry_SaveFlags.pUvctl = &uvctl;
  ICacheEntry_SaveFlags.pXyrange = &xyrange;
  ICacheEntry_SaveFlags.pUvrange = &uvrange;
}

void MPE::SaveRegisters()
{
  memcpy(tempScalarRegs, regs, sizeof(uint32) * 32);
  memcpy(&tempRc0, &rc0, sizeof(uint32) * 15);
}

void MPE::ExecuteNuances(InstructionCacheEntry &entry)
{
  uint32 i;

  if(entry.packetInfo & PACKETINFO_BREAKPOINT)
  {
    excepsrc |= 0x04;
    if(excephalten & 0x04)
    {
      //clear mpego bit
      mpectl &= ~MPECTRL_MPEGO;
    }
    else
    {
      //set exception bit in interrupt source register
      intsrc |= 0x01;
    }
  }

  if(!(entry.packetInfo & PACKETINFO_NOP))
  {
    if(entry.packetInfo & PACKETINFO_DEPENDENCY_PRESENT)
    {
      memcpy(tempScalarRegs, regs, sizeof(uint32) * 32);
      memcpy(&tempRc0, &rc0, sizeof(uint32) * 15);
    }

    tempCC = cc;

    for(i = 0; i < entry.nuanceCount; i++)
    {
      (nuanceHandlers[entry.handlers[i]])(*this,entry,*((Nuance *)(&entry.nuances[FIXED_FIELD(i,0)])));
    }
  }
}

uint8 MPE::DecodeSingleInstruction(uint8 *iPtr, InstructionCacheEntry *entry, uint32 *immExt, bool &bTerminating)
{
  //if 16 bit ALU instruction
  uint8 opcode = *iPtr;
  if(opcode <= 0x3F)
  {
    DecodeInstruction_ALU16(iPtr,entry,immExt);
    bTerminating = true;
    *immExt = 0;
    return 2;
  }
  else
  {
    //Instruction is not a 16 bit ALU instruction

    //If 16 bit non-terminating non-ALU instruction, control instruction
    //or 32 bit immediate extension
    if(opcode < 0x88 || opcode > 0x9F)
    {
      //If not a control instruction (PAD, NOP or BREAKPOINT)
      if((opcode & 0xFC) != 0x80)
      {
        if((opcode & 0x7F) <= 0x67)
        {
          if((opcode & 0x7F) >= 0x48)
          {
            DecodeInstruction_MEM16(iPtr,entry,immExt);
          }
          else
          {
            DecodeInstruction_MUL16(iPtr,entry,immExt);
          }
        }
        else
        {
          if((opcode & 0x7F) <= 0x73)
          {
            DecodeInstruction_ECU16(iPtr,entry,immExt);
          }
          else
          {
            DecodeInstruction_RCU16(iPtr,entry,immExt);
          }
        }

        //if bit 15 is set, the instruction is a packet terminator
        bTerminating = (opcode & 0x80);
      }
      else
      {
        //PAD, NOP, or BREAKPOINT.
        entry->packetInfo |= (opcode & 0x03);

        //if bit 8 is set, the instruction is a terminating breakpoint or NOP
        bTerminating = (opcode & 0x01);
      }

      *immExt = 0;
      return 2;
    }
    else if(opcode >= 0x90)
    {
      switch((opcode & 0x0C) >> 2)
      {
        case 0:
          DecodeInstruction_ECU32(iPtr,entry,immExt);
          break;
        case 1:
          DecodeInstruction_MEM32(iPtr,entry,immExt);
          break;
        case 2:
          DecodeInstruction_ALU32(iPtr,entry,immExt);
          break;
        case 3:
          DecodeInstruction_MUL32(iPtr,entry,immExt);
          break;
      }

      //32 bit instruction: if bit 12 is set, instruction is a packet terminator
      bTerminating = *(iPtr + 2) & 0x10;
      *immExt = 0;
      return 4;
    }
    else
    {
      //32 bit extension (preceeding 48/64 bit instruction)
      *immExt = *((uint32 *)iPtr);
      SwapScalarBytes(immExt);
      bTerminating = false;
      return 4;
    }
  }
  return 0;
}

NativeCodeCacheEntryPoint MPE::CompileNativeCodeBlock(uint32 pcexec, SuperBlockCompileType compileType, bool &bError, bool bSinglePacket)
{
  return superBlock->CompileBlock(this, pcexec, *nativeCodeCache, compileType, bSinglePacket, bError);
}

void MPE::PrintInstructionCachePacket(char *buffer, InstructionCacheEntry &entry)
{
  if(entry.nuanceCount != 0)
  {
    for(uint32 i = 0; i < entry.nuanceCount; i++)
    {
      buffer += (printHandlers[entry.handlers[i]])(buffer, *((Nuance *)(&entry.nuances[FIXED_FIELD(i,0)])), true);
    }
  }
  
  if(entry.packetInfo & PACKETINFO_BREAKPOINT)
  {
      sprintf(buffer,"breakpoint\n");
  }
  else if(entry.packetInfo & PACKETINFO_NOP)
  {
    sprintf(buffer,"nop\n");
  }
}

void MPE::PrintInstructionCachePacket(char *buffer, uint32 address)
{
  InstructionCacheEntry entry;
  uint8 *memPtr;
  uint32 bCacheEntryValid;

  bCacheEntryValid = 1;
  InstructionCacheEntry *pEntry = instructionCache->FindInstructionCacheEntry(address,bCacheEntryValid);
  
  if(bCacheEntryValid && (address == pEntry->pcexec))
  {
    PrintInstructionCachePacket(buffer,*pEntry);
  }
  else
  {
    entry.pcexec = address;
    memPtr = GetPointerToMemoryBank(address);
    DecompressPacket(memPtr,&entry);
    PrintInstructionCachePacket(buffer,entry);    
  }
}

void MPE::ExecuteSingleStep()
{
  InvalidateICacheRegion(pcexec, pcexec);
  nativeCodeCache->FlushRegion(pcexec, pcexec);
  FetchDecodeExecute();
}